use crate::value::Value;
use tokio::time::Instant;

#[derive(Debug)]
pub struct Entry {
    pub value: Value,
    expires_at: Option<Instant>,
}

/// Database Entry
///
/// A database entry is a Value associated with an optional ttl.
///
/// The database will never return an entry if has expired already, by having
/// this promise we can run the purge process every few seconds instead of doing
/// so more frequently.
impl Entry {
    pub fn new(value: Value, expires_at: Option<Instant>) -> Self {
        Self { value, expires_at }
    }

    pub fn persist(&mut self) {
        self.expires_at = None;
    }

    pub fn get_ttl(&self) -> Option<Instant> {
        self.expires_at
    }

    pub fn has_ttl(&self) -> bool {
        self.expires_at.is_some()
    }

    pub fn set_ttl(&mut self, expires_at: Instant) {
        self.expires_at = Some(expires_at);
    }

    /// Changes the value that is wrapped in this entry, the TTL (expired_at) is
    /// not affected.
    pub fn change_value(&mut self, value: Value) {
        self.value = value;
    }

    #[allow(dead_code)]
    pub fn get_mut(&mut self) -> &mut Value {
        &mut self.value
    }

    pub fn get(&self) -> &Value {
        &self.value
    }

    /// If the Entry should be taken as valid, if this function returns FALSE
    /// the callee should behave as if the key was not found. By having this
    /// behaviour we can schedule the purge thread to run every few seconds or
    /// even minutes instead of once every second.
    pub fn is_valid(&self) -> bool {
        self.expires_at.map_or(true, |x| x > Instant::now())
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use tokio::time::Duration;

    #[test]
    fn is_valid_without_expiration() {
        let e = Entry::new(Value::Null, None);
        assert!(e.is_valid());
    }

    #[test]
    fn is_valid() {
        let e = (
            Entry::new(Value::Null, Some(Instant::now() - Duration::from_secs(5))),
            Entry::new(Value::Null, Some(Instant::now())),
            Entry::new(Value::Null, Some(Instant::now() + Duration::from_secs(5))),
        );
        assert!(!e.0.is_valid());
        assert!(!e.1.is_valid());
        assert!(e.2.is_valid());
    }

    #[test]
    fn persist() {
        let mut e = Entry::new(Value::Null, Some(Instant::now()));
        assert!(!e.is_valid());
        e.persist();
        assert!(e.is_valid());
    }

    #[test]
    fn update_ttl() {
        let mut e = Entry::new(Value::Null, Some(Instant::now()));
        assert!(!e.is_valid());
        e.persist();
        assert!(e.is_valid());
        e.set_ttl(Instant::now());
        assert!(!e.is_valid());
    }
}
